@@ -1,32 +1,17 @@
-%%   The contents of this file are subject to the Mozilla Public License
-%%   Version 1.1 (the "License"); you may not use this file except in
-%%   compliance with the License. You may obtain a copy of the License at
-%%   http://www.mozilla.org/MPL/
+%% The contents of this file are subject to the Mozilla Public License
+%% Version 1.1 (the "License"); you may not use this file except in
+%% compliance with the License. You may obtain a copy of the License
+%% at http://www.mozilla.org/MPL/
 %%
-%%   Software distributed under the License is distributed on an "AS IS"
-%%   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
-%%   License for the specific language governing rights and limitations
-%%   under the License.
+%% Software distributed under the License is distributed on an "AS IS"
+%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
+%% the License for the specific language governing rights and
+%% limitations under the License.
 %%
-%%   The Original Code is RabbitMQ.
+%% The Original Code is RabbitMQ.
 %%
-%%   The Initial Developers of the Original Code are LShift Ltd,
-%%   Cohesive Financial Technologies LLC, and Rabbit Technologies Ltd.
-%%
-%%   Portions created before 22-Nov-2008 00:00:00 GMT by LShift Ltd,
-%%   Cohesive Financial Technologies LLC, or Rabbit Technologies Ltd
-%%   are Copyright (C) 2007-2008 LShift Ltd, Cohesive Financial
-%%   Technologies LLC, and Rabbit Technologies Ltd.
-%%
-%%   Portions created by LShift Ltd are Copyright (C) 2007-2010 LShift
-%%   Ltd. Portions created by Cohesive Financial Technologies LLC are
-%%   Copyright (C) 2007-2010 Cohesive Financial Technologies
-%%   LLC. Portions created by Rabbit Technologies Ltd are Copyright
-%%   (C) 2007-2010 Rabbit Technologies Ltd.
-%%
-%%   All Rights Reserved.
-%%
-%%   Contributor(s): ______________________________________.
+%% The Initial Developer of the Original Code is VMware, Inc.
+%% Copyright (c) 2007-2012 VMware, Inc.  All rights reserved.
 %%
 
 -module(rabbit_auth_backend_ldap).
@@ -52,6 +37,8 @@
 
 -record(state, { servers,
                  user_dn_pattern,
+                 dn_lookup_attribute,
+                 dn_lookup_base,
                  other_bind,
                  vhost_access_query,
                  resource_access_query,
@@ -60,6 +47,8 @@
                  log,
                  port }).
 
+-record(impl, { user_dn, password }).
+
 %%--------------------------------------------------------------------
 
 description() ->
@@ -81,13 +70,14 @@ check_user_login(Username, AuthProps) ->
     exit({unknown_auth_props, Username, AuthProps}).
 
 check_vhost_access(User = #user{username = Username,
-                                impl = UserDN}, VHost) ->
+                                impl     = #impl{user_dn = UserDN}}, VHost) ->
     gen_server:call(?SERVER, {check_vhost, [{username, Username},
                                             {user_dn,  UserDN},
                                             {vhost,    VHost}], User},
                     infinity).
 
-check_resource_access(User = #user{username = Username, impl = UserDN},
+check_resource_access(User = #user{username = Username,
+                                   impl     = #impl{user_dn = UserDN}},
                       #resource{virtual_host = VHost, kind = Type, name = Name},
                       Permission) ->
     gen_server:call(?SERVER, {check_resource, [{username,   Username},
@@ -118,7 +108,8 @@ evaluate({exists, DNPattern}, Args, _User, LDAP) ->
     Filter = eldap:present("objectClass"),
     object_exists(DNPattern, Filter, Args, LDAP);
 
-evaluate({in_group, DNPattern}, Args, #user{impl = UserDN}, LDAP) ->
+evaluate({in_group, DNPattern}, Args, #user{impl = #impl{user_dn = UserDN}},
+         LDAP) ->
     Filter = eldap:equalityMatch("member", UserDN),
     object_exists(DNPattern, Filter, Args, LDAP);
 
@@ -143,11 +134,12 @@ object_exists(DNPattern, Filter, Args, LDAP) ->
     case eldap:search(LDAP,
                       [{base, DN},
                        {filter, Filter},
+                       {attributes, ["objectClass"]}, %% Reduce verbiage
                        {scope, eldap:baseObject()}]) of
         {ok, #eldap_search_result{entries = Entries}} ->
             length(Entries) > 0;
-        {error, _} ->
-            false
+        {error, _} = E ->
+            E
     end.
 
 attribute(DNPattern, AttributeName, Args, LDAP) ->
@@ -161,25 +153,22 @@ attribute(DNPattern, AttributeName, Args, LDAP) ->
             Attr;
         {ok, #eldap_search_result{entries = _}} ->
             false;
-        {error, _} ->
-            false
+        {error, _} = E ->
+            E
     end.
 
 evaluate_ldap(Q, Args, User, State) ->
-    with_ldap(fun(LDAP) -> evaluate(Q, Args, User, LDAP) end, State).
+    with_ldap(creds(User, State),
+              fun(LDAP) -> evaluate(Q, Args, User, LDAP) end, State).
 
 %%--------------------------------------------------------------------
 
-with_ldap(Fun, State = #state{ other_bind = BindOpts }) ->
-    with_ldap(BindOpts, Fun, State).
-
 %% TODO - ATM we create and destroy a new LDAP connection on every
 %% call. This could almost certainly be more efficient.
-with_ldap(BindOpts, Fun,
-          State = #state{ servers = Servers,
-                          use_ssl = SSL,
-                          log     = Log,
-                          port    = Port }) ->
+with_ldap(Creds, Fun, State = #state{servers = Servers,
+                                     use_ssl = SSL,
+                                     log     = Log,
+                                     port    = Port}) ->
     Opts0 = [{ssl, SSL}, {port, Port}],
     Opts = case Log of
                true ->
@@ -194,7 +183,7 @@ with_ldap(BindOpts, Fun,
     case eldap:open(Servers, Opts) of
         {ok, LDAP} ->
             Reply = try
-                        case BindOpts of
+                        case Creds of
                             anon ->
                                 Fun(LDAP);
                             {UserDN, Password} ->
@@ -220,11 +209,13 @@ get_env(F) ->
     {ok, V} = application:get_env(F),
     V.
 
-do_login(Username, LDAP, State = #state{ tag_queries = TagQueries }) ->
-    UserDN = username_to_dn(Username, State),
+do_login(Username, Password, LDAP,
+         State = #state{ tag_queries = TagQueries }) ->
+    UserDN = username_to_dn(Username, LDAP, State),
     User = #user{username     = Username,
                  auth_backend = ?MODULE,
-                 impl         = UserDN},
+                 impl         = #impl{user_dn  = UserDN,
+                                      password = Password}},
     TagRes = [{Tag, evaluate(Q, [{username, Username},
                                  {user_dn,  UserDN}], User, LDAP)} ||
                  {Tag, Q} <- TagQueries],
@@ -233,9 +224,38 @@ do_login(Username, LDAP, State = #state{ tag_queries = TagQueries }) ->
         [E | _] -> E
     end.
 
-username_to_dn(Username, #state{ user_dn_pattern = UserDNPattern }) ->
+username_to_dn(Username, _LDAP, State = #state{dn_lookup_attribute = none}) ->
+    fill_user_dn_pattern(Username, State);
+
+username_to_dn(Username, LDAP, State = #state{dn_lookup_attribute = Attr,
+                                              dn_lookup_base      = Base}) ->
+    Filled = fill_user_dn_pattern(Username, State),
+    case eldap:search(LDAP,
+                      [{base, Base},
+                       {filter, eldap:equalityMatch(Attr, Filled)},
+                       {attributes, ["distinguishedName"]}]) of
+        {ok, #eldap_search_result{entries = [#eldap_entry{attributes = A}]}} ->
+            [DN] = proplists:get_value("distinguishedName", A),
+            DN;
+        {ok, #eldap_search_result{entries = Entries}} ->
+            rabbit_log:warning("Searching for DN for ~s, got back ~p~n",
+                               [Filled, Entries]),
+            Filled;
+        {error, _} = E ->
+            exit(E)
+    end.
+
+fill_user_dn_pattern(Username, #state{user_dn_pattern = UserDNPattern}) ->
     rabbit_auth_backend_ldap_util:fill(UserDNPattern, [{username, Username}]).
 
+creds(none, #state{other_bind = as_user}) ->
+    exit(as_user_no_password);
+creds(#user{impl = #impl{user_dn = UserDN, password = Password}},
+      #state{other_bind = as_user}) ->
+    {UserDN, Password};
+creds(_, #state{other_bind = Creds}) ->
+    Creds.
+
 %%--------------------------------------------------------------------
 
 init([]) ->
@@ -244,11 +264,13 @@ init([]) ->
 
 handle_call({login, Username}, _From, State) ->
     %% Without password, e.g. EXTERNAL
-    with_ldap(fun(LDAP) -> do_login(Username, LDAP, State) end, State);
+    with_ldap(creds(none, State),
+              fun(LDAP) -> do_login(Username, none, LDAP, State) end, State);
 
 handle_call({login, Username, Password}, _From, State) ->
-    with_ldap({username_to_dn(Username, State), Password},
-              fun(LDAP) -> do_login(Username, LDAP, State) end, State);
+    with_ldap({fill_user_dn_pattern(Username, State), Password},
+              fun(LDAP) -> do_login(Username, Password, LDAP, State) end,
+              State);
 
 handle_call({check_vhost, Args, User},
             _From, State = #state{vhost_access_query = Q}) ->
