@@ -11,7 +11,7 @@
 %% The Original Code is RabbitMQ.
 %%
 %% The Initial Developer of the Original Code is VMware, Inc.
-%% Copyright (c) 2007-2011 VMware, Inc.  All rights reserved.
+%% Copyright (c) 2007-2012 VMware, Inc.  All rights reserved.
 %%
 
 %% @private
@@ -245,7 +245,7 @@ client_properties(UserProperties) ->
                {<<"version">>,   longstr, list_to_binary(Vsn)},
                {<<"platform">>,  longstr, <<"Erlang">>},
                {<<"copyright">>, longstr,
-                <<"Copyright (c) 2007-2011 VMware, Inc.">>},
+                <<"Copyright (c) 2007-2012 VMware, Inc.">>},
                {<<"information">>, longstr,
                 <<"Licensed under the MPL.  "
                   "See http://www.rabbitmq.com/">>},
@@ -256,7 +256,7 @@ client_properties(UserProperties) ->
 
 handshake_recv(Expecting) ->
     receive
-        {'$gen_cast', {method, Method, none}} ->
+        {'$gen_cast', {method, Method, none, noflow}} ->
             case {Expecting, element(1, Method)} of
                 {E, M} when E =:= M ->
                     Method;
