@@ -21,10 +21,12 @@ var SORT_QUERIES  = map(['/connections', '/channels', '/vhosts', '/users',
                          '/queues', '/exchanges']);
 
 // Extension arguments that we know about and present specially in the UI.
-var KNOWN_ARGS = {'alternate-exchange': {'short': 'AE',  'type': 'string'},
-                  'x-message-ttl':      {'short': 'TTL', 'type': 'int'},
-                  'x-expires':          {'short': 'Exp', 'type': 'int'},
-                  'x-ha-policy':        {'short': 'HA',  'type': 'string'}};
+var KNOWN_ARGS = {'alternate-exchange':        {'short': 'AE',  'type': 'string'},
+                  'x-message-ttl':             {'short': 'TTL', 'type': 'int'},
+                  'x-expires':                 {'short': 'Exp', 'type': 'int'},
+                  'x-ha-policy':               {'short': 'HA',  'type': 'string'},
+                  'x-dead-letter-exchange':    {'short': 'DLX', 'type': 'string'},
+                  'x-dead-letter-routing-key': {'short': 'DLK', 'type': 'string'}};
 
 // Things that are like arguments that we format the same way in listings.
 var IMPLICIT_ARGS = {'durable':         {'short': 'D',   'type': 'boolean'},
@@ -59,19 +61,27 @@ var extension_count;
 // The dispatcher needs access to the Sammy app
 var app;
 
+var exchange_types;
+
 // Set up the above vars
 function setup_global_vars() {
-    statistics_level = JSON.parse(sync_get('/overview')).statistics_level;
+    var overview = JSON.parse(sync_get('/overview'));
+    var nodes = JSON.parse(sync_get('/nodes'));
+    statistics_level = overview.statistics_level;
     var user = JSON.parse(sync_get('/whoami'));
     replace_content('login', '<p>User: <b>' + user.name + '</b></p>');
     var tags = user.tags.split(",");
     user_administrator = jQuery.inArray("administrator", tags) != -1;
     user_monitor = user_administrator ||
         jQuery.inArray("monitoring", tags) != -1;
-    nodes_interesting = user_monitor &&
-        JSON.parse(sync_get('/nodes')).length > 1;
+    nodes_interesting = user_monitor && nodes.length > 1;
     vhosts_interesting = JSON.parse(sync_get('/vhosts')).length > 1;
     current_vhost = get_pref('vhost');
+    for (var i in nodes) {
+        if (nodes[i].name == overview.node) {
+            exchange_types = nodes[i].exchange_types;
+        }
+    }
 }
 
 ////////////////////////////////////////////////////
