@@ -17,6 +17,12 @@ HELP = {
     'queue-auto-delete':
       'If yes, the queue will delete itself after at least one consumer has connected, and then all consumers have disconnected.',
 
+    'queue-dead-letter-exchange':
+      'Optional name of an exchange to which messages will be republished if they are rejected or expire.<br/>(Sets the "x-dead-letter-exchange" argument.)',
+
+    'queue-dead-letter-routing-key':
+      'Optional replacement routing key to use when a message is dead-lettered. If this is not set, the message\'s original routing key will be used.<br/>(Sets the "x-dead-letter-routing-key" argument.)',
+
     'internal-users-only':
       'Only users within the internal RabbitMQ database are shown here. Other users (e.g. those authenticated over LDAP) will not appear.',
 
