@@ -1,37 +1,23 @@
-%%   The contents of this file are subject to the Mozilla Public License
-%%   Version 1.1 (the "License"); you may not use this file except in
-%%   compliance with the License. You may obtain a copy of the License at
-%%   http://www.mozilla.org/MPL/
+%% The contents of this file are subject to the Mozilla Public License
+%% Version 1.1 (the "License"); you may not use this file except in
+%% compliance with the License. You may obtain a copy of the License
+%% at http://www.mozilla.org/MPL/
 %%
-%%   Software distributed under the License is distributed on an "AS IS"
-%%   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
-%%   License for the specific language governing rights and limitations
-%%   under the License.
+%% Software distributed under the License is distributed on an "AS IS"
+%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
+%% the License for the specific language governing rights and
+%% limitations under the License.
 %%
-%%   The Original Code is RabbitMQ.
+%% The Original Code is RabbitMQ.
 %%
-%%   The Initial Developers of the Original Code are LShift Ltd,
-%%   Cohesive Financial Technologies LLC, and Rabbit Technologies Ltd.
-%%
-%%   Portions created before 22-Nov-2008 00:00:00 GMT by LShift Ltd,
-%%   Cohesive Financial Technologies LLC, or Rabbit Technologies Ltd
-%%   are Copyright (C) 2007-2008 LShift Ltd, Cohesive Financial
-%%   Technologies LLC, and Rabbit Technologies Ltd.
-%%
-%%   Portions created by LShift Ltd are Copyright (C) 2007-2009 LShift
-%%   Ltd. Portions created by Cohesive Financial Technologies LLC are
-%%   Copyright (C) 2007-2009 Cohesive Financial Technologies
-%%   LLC. Portions created by Rabbit Technologies Ltd are Copyright
-%%   (C) 2007-2009 Rabbit Technologies Ltd.
-%%
-%%   All Rights Reserved.
-%%
-%%   Contributor(s): ______________________________________.
+%% The Initial Developer of the Original Code is VMware, Inc.
+%% Copyright (c) 2007-2012 VMware, Inc.  All rights reserved.
 %%
+
 -module(rabbit_stomp_processor).
 -behaviour(gen_server2).
 
--export([start_link/3, process_frame/2, flush_and_die/1]).
+-export([start_link/1, process_frame/2, flush_and_die/1]).
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
          code_change/3, terminate/2]).
 
@@ -41,10 +27,10 @@
 -include("rabbit_stomp_prefixes.hrl").
 -include("rabbit_stomp_headers.hrl").
 
--record(state, {socket, session_id, channel,
-                connection, subscriptions, version,
-                start_heartbeat_fun, pending_receipts,
-                config, reply_queues, frame_transformer}).
+-record(state, {session_id, channel, connection, subscriptions,
+                version, start_heartbeat_fun, pending_receipts,
+                config, dest_queues, reply_queues, frame_transformer,
+                adapter_info, send_fun, ssl_login_name}).
 
 -record(subscription, {dest_hdr, channel, multi_ack, description}).
 
@@ -54,12 +40,14 @@
 %%----------------------------------------------------------------------------
 %% Public API
 %%----------------------------------------------------------------------------
-start_link(Sock, StartHeartbeatFun, Configuration) ->
-    gen_server2:start_link(?MODULE, [Sock, StartHeartbeatFun, Configuration],
-                           []).
+start_link(Args) ->
+    gen_server2:start_link(?MODULE, Args, []).
 
+process_frame(Pid, Frame = #stomp_frame{command = "SEND"}) ->
+    credit_flow:send(Pid),
+    gen_server2:cast(Pid, {"SEND", Frame, self()});
 process_frame(Pid, Frame = #stomp_frame{command = Command}) ->
-    gen_server2:cast(Pid, {Command, Frame}).
+    gen_server2:cast(Pid, {Command, Frame, noflow}).
 
 flush_and_die(Pid) ->
     gen_server2:cast(Pid, flush_and_die).
@@ -68,11 +56,10 @@ flush_and_die(Pid) ->
 %% Basic gen_server2 callbacks
 %%----------------------------------------------------------------------------
 
-init([Sock, StartHeartbeatFun, Configuration]) ->
+init([SendFun, AdapterInfo, StartHeartbeatFun, SSLLoginName, Configuration]) ->
     process_flag(trap_exit, true),
     {ok,
      #state {
-       socket              = Sock,
        session_id          = none,
        channel             = none,
        connection          = none,
@@ -81,8 +68,12 @@ init([Sock, StartHeartbeatFun, Configuration]) ->
        start_heartbeat_fun = StartHeartbeatFun,
        pending_receipts    = undefined,
        config              = Configuration,
+       dest_queues         = sets:new(),
        reply_queues        = dict:new(),
-       frame_transformer   = undefined},
+       frame_transformer   = undefined,
+       adapter_info        = AdapterInfo,
+       send_fun            = SendFun,
+       ssl_login_name      = SSLLoginName},
      hibernate,
      {backoff, 1000, 1000, 10000}
     }.
@@ -93,18 +84,21 @@ terminate(_Reason, State) ->
 handle_cast(flush_and_die, State) ->
     {stop, normal, close_connection(State)};
 
-handle_cast({"STOMP", Frame}, State) ->
+handle_cast({"STOMP", Frame, noflow}, State) ->
     process_connect(no_implicit, Frame, State);
 
-handle_cast({"CONNECT", Frame}, State) ->
+handle_cast({"CONNECT", Frame, noflow}, State) ->
     process_connect(no_implicit, Frame, State);
 
 handle_cast(Request, State = #state{channel = none,
                                      config = #stomp_configuration{
                                       implicit_connect = true}}) ->
-    {noreply, State1, _} =
+    {noreply, State1 = #state{channel = Ch}, _} =
         process_connect(implicit, #stomp_frame{headers = []}, State),
-    handle_cast(Request, State1);
+    case Ch of
+        none -> {stop, normal, State1};
+        _    -> handle_cast(Request, State1)
+    end;
 
 handle_cast(_Request, State = #state{channel = none,
                                      config = #stomp_configuration{
@@ -115,7 +109,12 @@ handle_cast(_Request, State = #state{channel = none,
                 State),
      hibernate};
 
-handle_cast({Command, Frame}, State = #state{frame_transformer = FT}) ->
+handle_cast({Command, Frame, FlowPid},
+            State = #state{frame_transformer = FT}) ->
+    case FlowPid of
+        noflow -> ok;
+        _      -> credit_flow:ack(FlowPid)
+    end,
     Frame1 = FT(Frame),
     process_request(
       fun(StateN) ->
@@ -139,8 +138,19 @@ handle_info(#'basic.ack'{delivery_tag = Tag, multiple = IsMulti}, State) ->
 handle_info({Delivery = #'basic.deliver'{},
              #amqp_msg{props = Props, payload = Payload}}, State) ->
     {noreply, send_delivery(Delivery, Props, Payload, State), hibernate};
+handle_info({'EXIT', Conn,
+             {shutdown, {server_initiated_close, Code, Explanation}}},
+            State = #state{connection = Conn}) ->
+    amqp_death(Code, Explanation, State);
+handle_info({'EXIT', Conn, Reason}, State = #state{connection = Conn}) ->
+    send_error("AMQP connection died", "Reason: ~p", [Reason], State),
+    {stop, {conn_died, Reason}, State};
 handle_info({inet_reply, _, ok}, State) ->
     {noreply, State, hibernate};
+handle_info({bump_credit, Msg}, State) ->
+    credit_flow:handle_bump_msg(Msg),
+    {noreply, State, hibernate};
+
 handle_info({inet_reply, _, Status}, State) ->
     {stop, Status, State}.
 
@@ -171,39 +181,28 @@ process_request(ProcessFun, SuccessFun, State) ->
             {stop, R, NewState}
     end.
 
-process_connect(Implicit,
-                Frame,
-                State = #state{
-                  channel = none,
-                  socket  = Sock,
-                  config  = #stomp_configuration{
-                    default_login    = DefaultLogin,
-                    default_passcode = DefaultPasscode}}) ->
+process_connect(Implicit, Frame,
+                State = #state{channel        = none,
+                               config         = Config,
+                               ssl_login_name = SSLLoginName,
+                               adapter_info   = AdapterInfo}) ->
     process_request(
       fun(StateN) ->
               case negotiate_version(Frame) of
                   {ok, Version} ->
                       FT = frame_transformer(Version),
                       Frame1 = FT(Frame),
+                      {Username, Creds} = creds(Frame1, SSLLoginName, Config),
                       {ok, DefaultVHost} =
                           application:get_env(rabbit, default_vhost),
+                      {ProtoName, _} = AdapterInfo#adapter_info.protocol,
                       Res = do_login(
-                                rabbit_stomp_frame:header(Frame1,
-                                                          ?HEADER_LOGIN,
-                                                          DefaultLogin),
-                                rabbit_stomp_frame:header(Frame1,
-                                                          ?HEADER_PASSCODE,
-                                                          DefaultPasscode),
-                                rabbit_stomp_frame:header(Frame1,
-                                                          ?HEADER_HOST,
-                                                          binary_to_list(
-                                                            DefaultVHost)),
-                                rabbit_stomp_frame:header(Frame1,
-                                                          ?HEADER_HEART_BEAT,
-                                                          "0,0"),
-                                adapter_info(Sock, Version),
-                                Version,
-                                StateN#state{frame_transformer = FT}),
+                              Username, Creds,
+                              login_header(Frame1, ?HEADER_HOST, DefaultVHost),
+                              login_header(Frame1, ?HEADER_HEART_BEAT, "0,0"),
+                                AdapterInfo#adapter_info{
+                                  protocol = {ProtoName, Version}}, Version,
+                              StateN#state{frame_transformer = FT}),
                       case {Res, Implicit} of
                           {{ok, _, StateN1}, implicit} -> ok(StateN1);
                           _                            -> Res
@@ -218,6 +217,26 @@ process_connect(Implicit,
       fun(StateM) -> StateM end,
       State).
 
+creds(Frame, SSLLoginName,
+      #stomp_configuration{default_login    = DefLogin,
+                           default_passcode = DefPasscode}) ->
+    PasswordCreds = {login_header(Frame, ?HEADER_LOGIN, DefLogin),
+                     [{password, login_header(
+                                   Frame, ?HEADER_PASSCODE, DefPasscode)}]},
+    case {rabbit_stomp_frame:header(Frame, ?HEADER_LOGIN), SSLLoginName} of
+        {not_found, none}    -> PasswordCreds;
+        {not_found, SSLName} -> {SSLName, []};
+        _                    -> PasswordCreds
+    end.
+
+login_header(Frame, Key, Default) when is_binary(Default) ->
+    login_header(Frame, Key, binary_to_list(Default));
+login_header(Frame, Key, Default) ->
+    case rabbit_stomp_frame:header(Frame, Key, Default) of
+        undefined -> undefined;
+        Hdr       -> list_to_binary(Hdr)
+    end.
+
 %%----------------------------------------------------------------------------
 %% Frame Transformation
 %%----------------------------------------------------------------------------
@@ -257,8 +276,8 @@ handle_frame("UNSUBSCRIBE", Frame, State) ->
 
 handle_frame("SEND", Frame, State) ->
     without_headers(?HEADERS_NOT_ON_SEND, "SEND", Frame, State,
-        fun (Command, Frame, State) ->
-            with_destination("SEND", Frame, State, fun do_send/4)
+        fun (_Command, Frame1, State1) ->
+            with_destination("SEND", Frame1, State1, fun do_send/4)
         end);
 
 handle_frame("ACK", Frame, State) ->
@@ -321,6 +340,12 @@ ack_action(Command, Frame,
 %% Internal helpers for processing frames callbacks
 %%----------------------------------------------------------------------------
 
+cancel_subscription({error, invalid_prefix}, _Frame, State) ->
+    error("Invalid id",
+          "UNSUBSCRIBE 'id' may not start with ~s~n",
+          [?TEMP_QUEUE_ID_PREFIX],
+          State);
+
 cancel_subscription({error, _}, _Frame, State) ->
     error("Missing destination or id",
           "UNSUBSCRIBE must include a 'destination' or 'id' header\n",
@@ -423,90 +448,54 @@ without_headers([], Command, Frame, State, Fun) ->
 
 do_login(undefined, _, _, _, _, _, State) ->
     error("Bad CONNECT", "Missing login or passcode header(s)\n", State);
-
-do_login(Username0, Password0, VirtualHost0, Heartbeat, AdapterInfo,
-         Version, State) ->
-    Username = list_to_binary(Username0),
-    Password = list_to_binary(Password0),
-    VirtualHost = list_to_binary(VirtualHost0),
-    case rabbit_access_control:check_user_pass_login(Username, Password) of
+do_login(Username, Creds, VirtualHost, Heartbeat, AdapterInfo, Version,
+         State) ->
+    case rabbit_access_control:check_user_login(Username, Creds) of
         {ok, _User} ->
             case amqp_connection:start(
                    #amqp_params_direct{username     = Username,
                                        virtual_host = VirtualHost,
                                        adapter_info = AdapterInfo}) of
                 {ok, Connection} ->
+                    link(Connection),
                     {ok, Channel} = amqp_connection:open_channel(Connection),
-                    SessionId = rabbit_guid:string_guid("session"),
+                    SessionId =
+                        rabbit_guid:string(rabbit_guid:gen_secure(), "session"),
                     {{SendTimeout, ReceiveTimeout}, State1} =
                         ensure_heartbeats(Heartbeat, State),
                     ok("CONNECTED",
                        [{?HEADER_SESSION, SessionId},
                         {?HEADER_HEART_BEAT,
                          io_lib:format("~B,~B", [SendTimeout, ReceiveTimeout])},
+                        {?HEADER_SERVER, server_header()},
                         {?HEADER_VERSION, Version}],
                        "",
                        State1#state{session_id = SessionId,
                                     channel    = Channel,
                                     connection = Connection});
                 {error, auth_failure} ->
+                    rabbit_log:error("STOMP login failed - auth_failure "
+                                     "(user vanished)~n"),
                     error("Bad CONNECT", "Authentication failure\n", State);
                 {error, access_refused} ->
+                    rabbit_log:warning("STOMP login failed - access_refused "
+                                       "(vhost access not allowed)~n"),
                     error("Bad CONNECT", "Authentication failure\n", State)
             end;
-        {refused, _Msg, _Args} ->
+        {refused, Msg, Args} ->
+            rabbit_log:warning("STOMP login failed: " ++ Msg ++ "\n", Args),
             error("Bad CONNECT", "Authentication failure\n", State)
     end.
 
-adapter_info(Sock, Version) ->
-    {Addr, Port} = case rabbit_net:sockname(Sock) of
-                       {ok, Res} -> Res;
-                       _         -> {unknown, unknown}
-                   end,
-    {PeerAddr, PeerPort} = case rabbit_net:peername(Sock) of
-                               {ok, Res2} -> Res2;
-                               _          -> {unknown, unknown}
-                           end,
-    #adapter_info{protocol        = {'STOMP', Version},
-                  address         = Addr,
-                  port            = Port,
-                  peer_address    = PeerAddr,
-                  peer_port       = PeerPort,
-                  additional_info = maybe_ssl_info(Sock)}.
-
-maybe_ssl_info(Sock) ->
-    case rabbit_net:is_ssl(Sock) of
-        true  -> [{ssl, true}] ++ ssl_info(Sock) ++ ssl_cert_info(Sock);
-        false -> [{ssl, false}]
-    end.
-
-ssl_info(Sock) ->
-    {Protocol, KeyExchange, Cipher, Hash} =
-        case rabbit_net:ssl_info(Sock) of
-            {ok, {P, {K, C, H}}}    -> {P, K, C, H};
-            {ok, {P, {K, C, H, _}}} -> {P, K, C, H};
-            _                       -> {unknown, unknown, unknown, unknown}
-        end,
-    [{ssl_protocol,       Protocol},
-     {ssl_key_exchange,   KeyExchange},
-     {ssl_cipher,         Cipher},
-     {ssl_hash,           Hash}].
-
-ssl_cert_info(Sock) ->
-    case rabbit_net:peercert(Sock) of
-        {ok, Cert} ->
-            [{peer_cert_issuer,   list_to_binary(
-                                    rabbit_ssl:peer_cert_issuer(Cert))},
-             {peer_cert_subject,  list_to_binary(
-                                    rabbit_ssl:peer_cert_subject(Cert))},
-             {peer_cert_validity, list_to_binary(
-                                    rabbit_ssl:peer_cert_validity(Cert))}];
-        _ ->
-            []
-    end.
+server_header() ->
+    Props = rabbit_reader:server_properties(?PROTOCOL),
+    {_, Product} = rabbit_misc:table_lookup(Props, <<"product">>),
+    {_, Version} = rabbit_misc:table_lookup(Props, <<"version">>),
+    rabbit_misc:format("~s/~s", [Product, Version]).
 
 do_subscribe(Destination, DestHdr, Frame,
              State = #state{subscriptions = Subs,
+                            dest_queues   = DestQs,
                             connection    = Connection,
                             channel       = MainChannel}) ->
     Prefetch =
@@ -526,7 +515,8 @@ do_subscribe(Destination, DestHdr, Frame,
 
     {AckMode, IsMulti} = rabbit_stomp_util:ack_mode(Frame),
 
-    {ok, Queue} = ensure_queue(subscribe, Destination, Frame, Channel),
+    {ok, Queue, DestQs1} = ensure_queue(subscribe, Destination, Frame, Channel,
+                                        DestQs),
 
     {ok, ConsumerTag, Description} = rabbit_stomp_util:consumer_tag(Frame),
 
@@ -547,14 +537,16 @@ do_subscribe(Destination, DestHdr, Frame,
                                                 channel     = Channel,
                                                 multi_ack   = IsMulti,
                                                 description = Description},
-                                  Subs)}).
+                                  Subs),
+                   dest_queues = DestQs1}).
 
 do_send(Destination, _DestHdr,
         Frame = #stomp_frame{body_iolist = BodyFragments},
-        State = #state{channel = Channel}) ->
-    {ok, _Q} = ensure_queue(send, Destination, Frame, Channel),
+        State = #state{channel = Channel, dest_queues = DestQs}) ->
+    {ok, _Q, DestQs1} = ensure_queue(send, Destination, Frame, Channel, DestQs),
 
-    {Frame1, State1} = ensure_reply_to(Frame, State),
+    {Frame1, State1} =
+        ensure_reply_to(Frame, State#state{dest_queues = DestQs1}),
 
     Props = rabbit_stomp_util:message_properties(Frame1),
 
@@ -625,8 +617,9 @@ send_method(Method, Properties, BodyFragments,
             State = #state{channel = Channel}) ->
     send_method(Method, Channel, Properties, BodyFragments, State).
 
-send_method(Method, Channel, Properties, BodyFragments, State) ->
-    amqp_channel:call(
+send_method(Method = #'basic.publish'{}, Channel, Properties, BodyFragments,
+            State) ->
+    amqp_channel:cast_flow(
       Channel, Method,
       #amqp_msg{props   = Properties,
                 payload = list_to_binary(BodyFragments)}),
@@ -673,12 +666,14 @@ ensure_reply_queue(TempQueueId, State = #state{channel       = Channel,
                                   #'queue.declare'{auto_delete = true,
                                                    exclusive   = true}),
 
-            #'basic.consume_ok'{consumer_tag = ConsumerTag} =
+            ConsumerTag = rabbit_stomp_util:consumer_tag_reply_to(TempQueueId),
+            #'basic.consume_ok'{} =
                 amqp_channel:subscribe(Channel,
                                        #'basic.consume'{
-                                         queue  = Queue,
-                                         no_ack = true,
-                                         nowait = false},
+                                         queue        = Queue,
+                                         consumer_tag = ConsumerTag,
+                                         no_ack       = true,
+                                         nowait       = false},
                                        self()),
 
             Destination = reply_to_destination(Queue),
@@ -847,18 +842,19 @@ perform_transaction_action({Method, Props, BodyFragments}, State) ->
 %%--------------------------------------------------------------------
 
 ensure_heartbeats(Heartbeats,
-                  State = #state{socket = Sock, start_heartbeat_fun = SHF}) ->
+                  State = #state{start_heartbeat_fun = SHF,
+                                 send_fun            = RawSendFun}) ->
     [CX, CY] = [list_to_integer(X) ||
                    X <- re:split(Heartbeats, ",", [{return, list}])],
 
-    SendFun = fun() -> catch rabbit_net:send(Sock, <<$\n>>) end,
+    SendFun = fun() -> RawSendFun(sync, <<$\n>>) end,
     Pid = self(),
     ReceiveFun = fun() -> gen_server2:cast(Pid, client_timeout) end,
 
     {SendTimeout, ReceiveTimeout} =
         {millis_to_seconds(CY), millis_to_seconds(CX)},
 
-    SHF(Sock, SendTimeout, SendFun, ReceiveTimeout, ReceiveFun),
+    SHF(SendTimeout, SendFun, ReceiveTimeout, ReceiveFun),
 
     {{SendTimeout * 1000 , ReceiveTimeout * 1000}, State}.
 
@@ -870,24 +866,29 @@ millis_to_seconds(M)               -> M div 1000.
 %% Queue and Binding Setup
 %%----------------------------------------------------------------------------
 
-ensure_queue(subscribe, {exchange, _}, _Frame, Channel) ->
+ensure_queue(subscribe, {exchange, _}, _Frame, Channel, DestQs) ->
     %% Create anonymous, exclusive queue for SUBSCRIBE on /exchange destinations
     #'queue.declare_ok'{queue = Queue} =
         amqp_channel:call(Channel, #'queue.declare'{auto_delete = true,
                                                     exclusive = true}),
-    {ok, Queue};
-ensure_queue(send, {exchange, _}, _Frame, _Channel) ->
+    {ok, Queue, DestQs};
+ensure_queue(send, {exchange, _}, _Frame, _Channel, DestQs) ->
     %% Don't create queues on SEND for /exchange destinations
-    {ok, undefined};
-ensure_queue(_, {queue, Name}, _Frame, Channel) ->
-    %% Always create named queue for /queue destinations
+    {ok, undefined, DestQs};
+ensure_queue(_, {queue, Name}, _Frame, Channel, DestQs) ->
+    %% Always create named queue for /queue destinations the first
+    %% time we encounter it
     Queue = list_to_binary(Name),
-    amqp_channel:cast(Channel,
-                      #'queue.declare'{durable = true,
-                                       queue   = Queue,
-                                       nowait  = true}),
-    {ok, Queue};
-ensure_queue(subscribe, {topic, Name}, Frame, Channel) ->
+    DestQs1 = case sets:is_element(Queue, DestQs) of
+                  true  -> DestQs;
+                  false -> amqp_channel:cast(Channel,
+                                             #'queue.declare'{durable = true,
+                                                              queue   = Queue,
+                                                              nowait  = true}),
+                           sets:add_element(Queue, DestQs)
+              end,
+    {ok, Queue, DestQs1};
+ensure_queue(subscribe, {topic, Name}, Frame, Channel, DestQs) ->
     %% Create queue for SUBSCRIBE on /topic destinations Queues are
     %% anonymous, auto_delete and exclusive for transient
     %% subscriptions. Durable subscriptions get shared, named, durable
@@ -905,13 +906,13 @@ ensure_queue(subscribe, {topic, Name}, Frame, Channel) ->
 
     #'queue.declare_ok'{queue = Queue} =
         amqp_channel:call(Channel, Method),
-    {ok, Queue};
-ensure_queue(send, {topic, _}, _Frame, _Channel) ->
+    {ok, Queue, DestQs};
+ensure_queue(send, {topic, _}, _Frame, _Channel, DestQs) ->
     %% Don't create queues on SEND for /topic destinations
-    {ok, undefined};
-ensure_queue(_, {Type, Name}, _Frame, _Channel)
+    {ok, undefined, DestQs};
+ensure_queue(_, {Type, Name}, _Frame, _Channel, DestQs)
   when Type =:= reply_queue orelse Type =:= amqqueue ->
-    {ok, list_to_binary(Name)}.
+    {ok, list_to_binary(Name), DestQs}.
 
 ensure_queue_binding(QueueBin, {"", Queue}, _Channel) ->
     %% i.e., we should only be asked to bind to the default exchange a
@@ -940,10 +941,9 @@ ok(Command, Headers, BodyFragments, State) ->
 
 amqp_death(ReplyCode, Explanation, State) ->
     ErrorName = ?PROTOCOL:amqp_exception(ReplyCode),
-    {stop, amqp_death,
-     send_error(atom_to_list(ErrorName),
-                format_detail("~s~n", [Explanation]),
-                State)}.
+    ErrorDesc = rabbit_misc:format("~s~n", [Explanation]),
+    log_error(ErrorName, ErrorDesc, none),
+    {stop, normal, send_error(atom_to_list(ErrorName), ErrorDesc, State)}.
 
 error(Message, Detail, State) ->
     priv_error(Message, Detail, none, State).
@@ -952,19 +952,20 @@ error(Message, Format, Args, State) ->
     priv_error(Message, Format, Args, none, State).
 
 priv_error(Message, Detail, ServerPrivateDetail, State) ->
-    error_logger:error_msg("STOMP error frame sent:~n" ++
-                           "Message: ~p~n" ++
-                           "Detail: ~p~n" ++
-                           "Server private detail: ~p~n",
-                           [Message, Detail, ServerPrivateDetail]),
+    log_error(Message, Detail, ServerPrivateDetail),
     {error, Message, Detail, State}.
 
 priv_error(Message, Format, Args, ServerPrivateDetail, State) ->
-    priv_error(Message, format_detail(Format, Args),
-                    ServerPrivateDetail, State).
+    priv_error(Message, rabbit_misc:format(Format, Args), ServerPrivateDetail,
+               State).
+
+log_error(Message, Detail, ServerPrivateDetail) ->
+    rabbit_log:error("STOMP error frame sent:~n"
+                     "Message: ~p~n"
+                     "Detail: ~p~n"
+                     "Server private detail: ~p~n",
+                     [Message, Detail, ServerPrivateDetail]).
 
-format_detail(Format, Args) ->
-    lists:flatten(io_lib:format(Format, Args)).
 %%----------------------------------------------------------------------------
 %% Frame sending utilities
 %%----------------------------------------------------------------------------
@@ -974,11 +975,8 @@ send_frame(Command, Headers, BodyFragments, State) ->
                             body_iolist = BodyFragments},
                State).
 
-send_frame(Frame, State = #state{socket = Sock}) ->
-    %% We ignore certain errors here, as we will be receiving an
-    %% asynchronous notification of the same (or a related) fault
-    %% shortly anyway. See bug 21365.
-    catch rabbit_net:port_command(Sock, rabbit_stomp_frame:serialize(Frame)),
+send_frame(Frame, State = #state{send_fun = SendFun}) ->
+    SendFun(async, rabbit_stomp_frame:serialize(Frame)),
     State.
 
 send_error(Message, Detail, State) ->
@@ -988,7 +986,7 @@ send_error(Message, Detail, State) ->
                          Detail, State).
 
 send_error(Message, Format, Args, State) ->
-    send_error(Message, format_detail(Format, Args), State).
+    send_error(Message, rabbit_misc:format(Format, Args), State).
 
 %%----------------------------------------------------------------------------
 %% Skeleton gen_server2 callbacks
@@ -998,4 +996,3 @@ handle_call(_Msg, _From, State) ->
 
 code_change(_OldVsn, State, _Extra) ->
     {ok, State}.
-
