@@ -11,7 +11,7 @@
 %% The Original Code is RabbitMQ.
 %%
 %% The Initial Developer of the Original Code is VMware, Inc.
-%% Copyright (c) 2007-2011 VMware, Inc.  All rights reserved.
+%% Copyright (c) 2007-2012 VMware, Inc.  All rights reserved.
 %%
 
 -module(rabbit_guid).
@@ -19,7 +19,7 @@
 -behaviour(gen_server).
 
 -export([start_link/0]).
--export([guid/0, string_guid/1, binstring_guid/1]).
+-export([gen/0, gen_secure/0, string/2, binary/2]).
 
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
          code_change/3]).
@@ -38,9 +38,10 @@
 -type(guid() :: binary()).
 
 -spec(start_link/0 :: () -> rabbit_types:ok_pid_or_error()).
--spec(guid/0 :: () -> guid()).
--spec(string_guid/1 :: (any()) -> string()).
--spec(binstring_guid/1 :: (any()) -> binary()).
+-spec(gen/0 :: () -> guid()).
+-spec(gen_secure/0 :: () -> guid()).
+-spec(string/2 :: (guid(), any()) -> string()).
+-spec(binary/2 :: (guid(), any()) -> binary()).
 
 -endif.
 
@@ -65,11 +66,8 @@ update_disk_serial() ->
     end,
     Serial.
 
-%% generate a GUID.
-%%
-%% The id is only unique within a single cluster and as long as the
-%% serial store hasn't been deleted.
-guid() ->
+%% Generate an un-hashed guid.
+fresh() ->
     %% We don't use erlang:now() here because a) it may return
     %% duplicates when the system clock has been rewound prior to a
     %% restart, or ids were generated at a high rate (which causes
@@ -78,29 +76,74 @@ guid() ->
     %%
     %% A persisted serial number, the node, and a unique reference
     %% (per node incarnation) uniquely identifies a process in space
-    %% and time. We combine that with a process-local counter to give
-    %% us a GUID.
-    G = case get(guid) of
-            undefined -> Serial = gen_server:call(?SERVER, serial, infinity),
-                         {{Serial, node(), make_ref()}, 0};
+    %% and time.
+    Serial = gen_server:call(?SERVER, serial, infinity),
+    {Serial, node(), make_ref()}.
+
+advance_blocks({B1, B2, B3, B4}, I) ->
+    %% To produce a new set of blocks, we create a new 32bit block
+    %% hashing {B5, I}. The new hash is used as last block, and the
+    %% other three blocks are XORed with it.
+    %%
+    %% Doing this is convenient because it avoids cascading conflits,
+    %% while being very fast. The conflicts are avoided by propagating
+    %% the changes through all the blocks at each round by XORing, so
+    %% the only occasion in which a collision will take place is when
+    %% all 4 blocks are the same and the counter is the same.
+    %%
+    %% The range (2^32) is provided explicitly since phash uses 2^27
+    %% by default.
+    B5 = erlang:phash2({B1, I}, 4294967296),
+    {{(B2 bxor B5), (B3 bxor B5), (B4 bxor B5), B5}, I+1}.
+
+blocks_to_binary({B1, B2, B3, B4}) -> <<B1:32, B2:32, B3:32, B4:32>>.
+
+%% generate a GUID. This function should be used when performance is a
+%% priority and predictability is not an issue. Otherwise use
+%% gen_secure/0.
+gen() ->
+    %% We hash a fresh GUID with md5, split it in 4 blocks, and each
+    %% time we need a new guid we rotate them producing a new hash
+    %% with the aid of the counter. Look at the comments in
+    %% advance_blocks/2 for details.
+    {BS, I} = case get(guid) of
+                  undefined -> <<B1:32, B2:32, B3:32, B4:32>> =
+                                   erlang:md5(term_to_binary(fresh())),
+                               {{B1,B2,B3,B4}, 0};
+                  {BS0, I0} -> advance_blocks(BS0, I0)
+              end,
+    put(guid, {BS, I}),
+    blocks_to_binary(BS).
+
+%% generate a non-predictable GUID.
+%%
+%% The id is only unique within a single cluster and as long as the
+%% serial store hasn't been deleted.
+%%
+%% If you are not concerned with predictability, gen/0 is faster.
+gen_secure() ->
+    %% Here instead of hashing once we hash the GUID and the counter
+    %% each time, so that the GUID is not predictable.
+    G = case get(guid_secure) of
+            undefined -> {fresh(), 0};
             {S, I}    -> {S, I+1}
         end,
-    put(guid, G),
+    put(guid_secure, G),
     erlang:md5(term_to_binary(G)).
 
 %% generate a readable string representation of a GUID.
 %%
 %% employs base64url encoding, which is safer in more contexts than
 %% plain base64.
-string_guid(Prefix) ->
+string(G, Prefix) ->
     Prefix ++ "-" ++ lists:foldl(fun ($\+, Acc) -> [$\- | Acc];
                                      ($\/, Acc) -> [$\_ | Acc];
                                      ($\=, Acc) -> Acc;
                                      (Chr, Acc) -> [Chr | Acc]
-                                 end, [], base64:encode_to_string(guid())).
+                                 end, [], base64:encode_to_string(G)).
 
-binstring_guid(Prefix) ->
-    list_to_binary(string_guid(Prefix)).
+binary(G, Prefix) ->
+    list_to_binary(string(G, Prefix)).
 
 %%----------------------------------------------------------------------------
 
