@@ -11,7 +11,7 @@
 %% The Original Code is RabbitMQ.
 %%
 %% The Initial Developer of the Original Code is VMware, Inc.
-%% Copyright (c) 2010-2011 VMware, Inc.  All rights reserved.
+%% Copyright (c) 2010-2012 VMware, Inc.  All rights reserved.
 %%
 
 -module(rabbit_mirror_queue_slave).
@@ -90,7 +90,7 @@
                }).
 
 start_link(Q) ->
-    gen_server2:start_link(?MODULE, [Q], []).
+    gen_server2:start_link(?MODULE, Q, []).
 
 set_maximum_since_use(QPid, Age) ->
     gen_server2:cast(QPid, {set_maximum_since_use, Age}).
@@ -98,59 +98,64 @@ set_maximum_since_use(QPid, Age) ->
 info(QPid) ->
     gen_server2:call(QPid, info, infinity).
 
-init([#amqqueue { name = QueueName } = Q]) ->
-    process_flag(trap_exit, true), %% amqqueue_process traps exits too.
-    {ok, GM} = gm:start_link(QueueName, ?MODULE, [self()]),
-    receive {joined, GM} ->
-            ok
-    end,
+init(#amqqueue { name = QueueName } = Q) ->
     Self = self(),
     Node = node(),
-    {ok, MPid} =
-        rabbit_misc:execute_mnesia_transaction(
-          fun () ->
-                  [Q1 = #amqqueue { pid = QPid, slave_pids = MPids }] =
-                      mnesia:read({rabbit_queue, QueueName}),
-                  %% ASSERTION
-                  [] = [Pid || Pid <- [QPid | MPids], node(Pid) =:= Node],
-                  MPids1 = MPids ++ [Self],
-                  ok = rabbit_amqqueue:store_queue(
-                         Q1 #amqqueue { slave_pids = MPids1 }),
-                  {ok, QPid}
-          end),
-    erlang:monitor(process, MPid),
-    ok = file_handle_cache:register_callback(
-           rabbit_amqqueue, set_maximum_since_use, [Self]),
-    ok = rabbit_memory_monitor:register(
-           Self, {rabbit_amqqueue, set_ram_duration_target, [Self]}),
-    {ok, BQ} = application:get_env(backing_queue_module),
-    BQS = bq_init(BQ, Q, false),
-    State = #state { q                   = Q,
-                     gm                  = GM,
-                     master_pid          = MPid,
-                     backing_queue       = BQ,
-                     backing_queue_state = BQS,
-                     rate_timer_ref      = undefined,
-                     sync_timer_ref      = undefined,
+    case rabbit_misc:execute_mnesia_transaction(
+           fun () ->
+                   [Q1 = #amqqueue { pid = QPid, slave_pids = MPids }] =
+                       mnesia:read({rabbit_queue, QueueName}),
+                   case [Pid || Pid <- [QPid | MPids], node(Pid) =:= Node] of
+                       []     -> MPids1 = MPids ++ [Self],
+                                 ok = rabbit_amqqueue:store_queue(
+                                        Q1 #amqqueue { slave_pids = MPids1 }),
+                                 {new, QPid};
+                       [SPid] -> true = rabbit_misc:is_process_alive(SPid),
+                                 existing
+                   end
+           end) of
+        {new, MPid} ->
+            process_flag(trap_exit, true), %% amqqueue_process traps exits too.
+            {ok, GM} = gm:start_link(QueueName, ?MODULE, [self()]),
+            receive {joined, GM} ->
+                    ok
+            end,
+            erlang:monitor(process, MPid),
+            ok = file_handle_cache:register_callback(
+                   rabbit_amqqueue, set_maximum_since_use, [Self]),
+            ok = rabbit_memory_monitor:register(
+                   Self, {rabbit_amqqueue, set_ram_duration_target, [Self]}),
+            {ok, BQ} = application:get_env(backing_queue_module),
+            BQS = bq_init(BQ, Q, false),
+            State = #state { q                   = Q,
+                             gm                  = GM,
+                             master_pid          = MPid,
+                             backing_queue       = BQ,
+                             backing_queue_state = BQS,
+                             rate_timer_ref      = undefined,
+                             sync_timer_ref      = undefined,
 
-                     sender_queues       = dict:new(),
-                     msg_id_ack          = dict:new(),
-                     ack_num             = 0,
+                             sender_queues       = dict:new(),
+                             msg_id_ack          = dict:new(),
+                             ack_num             = 0,
 
-                     msg_id_status       = dict:new(),
-                     known_senders       = dict:new(),
+                             msg_id_status       = dict:new(),
+                             known_senders       = dict:new(),
 
-                     synchronised        = false
+                             synchronised        = false
                    },
-    rabbit_event:notify(queue_slave_created,
-                        infos(?CREATION_EVENT_KEYS, State)),
-    ok = gm:broadcast(GM, request_length),
-    {ok, State, hibernate,
-     {backoff, ?HIBERNATE_AFTER_MIN, ?HIBERNATE_AFTER_MIN, ?DESIRED_HIBERNATE}}.
-
-handle_call({deliver_immediately, Delivery = #delivery {}}, From, State) ->
-    %% Synchronous, "immediate" delivery mode
+            rabbit_event:notify(queue_slave_created,
+                                infos(?CREATION_EVENT_KEYS, State)),
+            ok = gm:broadcast(GM, request_length),
+            {ok, State, hibernate,
+             {backoff, ?HIBERNATE_AFTER_MIN, ?HIBERNATE_AFTER_MIN,
+              ?DESIRED_HIBERNATE}};
+        existing ->
+            ignore
+    end.
 
+handle_call({deliver, Delivery = #delivery { immediate = true }},
+            From, State) ->
     %% It is safe to reply 'false' here even if a) we've not seen the
     %% msg via gm, or b) the master dies before we receive the msg via
     %% gm. In the case of (a), we will eventually receive the msg via
@@ -166,8 +171,8 @@ handle_call({deliver_immediately, Delivery = #delivery {}}, From, State) ->
     gen_server2:reply(From, false), %% master may deliver it, not us
     noreply(maybe_enqueue_message(Delivery, false, State));
 
-handle_call({deliver, Delivery = #delivery {}}, From, State) ->
-    %% Synchronous, "mandatory" delivery mode
+handle_call({deliver, Delivery = #delivery { mandatory = true }},
+            From, State) ->
     gen_server2:reply(From, true), %% amqqueue throws away the result anyway
     noreply(maybe_enqueue_message(Delivery, true, State));
 
@@ -208,8 +213,12 @@ handle_cast({run_backing_queue, Mod, Fun}, State) ->
 handle_cast({gm, Instruction}, State) ->
     handle_process_result(process_instruction(Instruction, State));
 
-handle_cast({deliver, Delivery = #delivery {}}, State) ->
+handle_cast({deliver, Delivery = #delivery{sender = Sender}, Flow}, State) ->
     %% Asynchronous, non-"mandatory", non-"immediate" deliver mode.
+    case Flow of
+        flow   -> credit_flow:ack(Sender);
+        noflow -> ok
+    end,
     noreply(maybe_enqueue_message(Delivery, true, State));
 
 handle_cast({set_maximum_since_use, Age}, State) ->
@@ -250,6 +259,10 @@ handle_info({'DOWN', _MonitorRef, process, ChPid, _Reason}, State) ->
 handle_info({'EXIT', _Pid, Reason}, State) ->
     {stop, Reason, State};
 
+handle_info({bump_credit, Msg}, State) ->
+    credit_flow:handle_bump_msg(Msg),
+    noreply(State);
+
 handle_info(Msg, State) ->
     {stop, {unexpected_info, Msg}, State}.
 
@@ -417,7 +430,7 @@ confirm_messages(MsgIds, State = #state { msg_id_status = MS }) ->
                           Acc
                   end
           end, {gb_trees:empty(), MS}, MsgIds),
-    rabbit_misc:gb_trees_foreach(fun rabbit_channel:confirm/2, CMs),
+    rabbit_misc:gb_trees_foreach(fun rabbit_misc:confirm_to_sender/2, CMs),
     State #state { msg_id_status = MS1 }.
 
 handle_process_result({ok,   State}) -> noreply(State);
@@ -447,7 +460,7 @@ promote_me(From, #state { q                   = Q = #amqqueue { name = QName },
     %% Everything that we're monitoring, we need to ensure our new
     %% coordinator is monitoring.
 
-    MonitoringPids = [begin true = erlang:demonitor(MRef),
+    MonitoringPids = [begin put({ch_publisher, Pid}, MRef),
                             Pid
                       end || {Pid, MRef} <- dict:to_list(KS)],
     ok = rabbit_mirror_queue_coordinator:ensure_monitoring(
@@ -601,7 +614,8 @@ ensure_monitoring(ChPid, State = #state { known_senders = KS }) ->
 local_sender_death(ChPid, State = #state { known_senders = KS }) ->
     ok = case dict:is_key(ChPid, KS) of
              false -> ok;
-             true  -> confirm_sender_death(ChPid)
+             true  -> credit_flow:peer_down(ChPid),
+                      confirm_sender_death(ChPid)
          end,
     State.
 
@@ -651,7 +665,7 @@ maybe_enqueue_message(
         {ok, {confirmed, ChPid}} ->
             %% BQ has confirmed it but we didn't know what the
             %% msg_seq_no was at the time. We do now!
-            ok = rabbit_channel:confirm(ChPid, [MsgSeqNo]),
+            ok = rabbit_misc:confirm_to_sender(ChPid, [MsgSeqNo]),
             SQ1 = remove_from_pending_ch(MsgId, ChPid, SQ),
             State1 #state { sender_queues = SQ1,
                             msg_id_status = dict:erase(MsgId, MS) };
@@ -668,7 +682,7 @@ maybe_enqueue_message(
                       msg_id_status =
                           dict:store(MsgId, {published, ChPid, MsgSeqNo}, MS) };
                 immediately ->
-                    ok = rabbit_channel:confirm(ChPid, [MsgSeqNo]),
+                    ok = rabbit_misc:confirm_to_sender(ChPid, [MsgSeqNo]),
                     SQ1 = remove_from_pending_ch(MsgId, ChPid, SQ),
                     State1 #state { msg_id_status = dict:erase(MsgId, MS),
                                     sender_queues = SQ1 }
@@ -730,7 +744,7 @@ process_instruction(
                         {MQ2, PendingCh,
                          dict:store(MsgId, {published, ChPid, MsgSeqNo}, MS)};
                     immediately ->
-                        ok = rabbit_channel:confirm(ChPid, [MsgSeqNo]),
+                        ok = rabbit_misc:confirm_to_sender(ChPid, [MsgSeqNo]),
                         {MQ2, PendingCh, MS}
                 end;
             {{value, {#delivery {}, _EnqueueOnPromotion}}, _MQ2} ->
@@ -829,6 +843,11 @@ process_instruction({ack, MsgIds},
     [] = MsgIds1 -- MsgIds, %% ASSERTION
     {ok, State #state { msg_id_ack          = MA1,
                         backing_queue_state = BQS1 }};
+process_instruction({fold, MsgFun, AckTags},
+                    State = #state { backing_queue       = BQ,
+                                     backing_queue_state = BQS }) ->
+    BQS1 = BQ:fold(AckTags, MsgFun, BQS),
+    {ok, State #state { backing_queue_state = BQS1 }};
 process_instruction({requeue, MsgIds},
                     State = #state { backing_queue       = BQ,
                                      backing_queue_state = BQS,
